/*
 * Copyright (c) 2001 Mark Fullmer and The Ohio State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      $Id: flow-fanout.c,v 1.8 2001/02/20 04:43:43 maf Exp $
 */

#if HAVE_CONFIG_H
 #include <config.h>
#endif

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/uio.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#if HAVE_STRINGS_H
 #include <strings.h>
#endif
#if HAVE_STRING_H
  #include <string.h>
#endif
#include <sys/time.h>
#include "support.h"

#define DEFAULT_PORT 9991

int debug;

void usage();

struct peer {
  int fd;
  struct sockaddr_in addr;
  int port;
};

int main(argc, argv)
int argc;
char **argv;
{
  extern char *optarg;
  extern int errno, optind;
  fd_set rfd;
  struct sockaddr_in fsocaddr;
  int fsockfd;
  int fport, i, len, n;
  int npeers, tx_delay;
  struct peer *peers;
  char tmp_buf[256];
  char rcv_buf[1500];
  char *c;

  fport = DEFAULT_PORT;
  debug = 0;
  tx_delay = 0;

  while ((i = getopt(argc, argv, "d:p:x:")) != -1)
    switch (i) {

    case 'd': /* debug */
      debug = atoi(optarg);
      break;

    case 'p': /* flow Port */
      fport = atoi(optarg);
      break;

    case 'x': /* transmit delay */
      tx_delay = atoi(optarg);
      break;

    default:
      usage();
      exit (1);
      break;

    } /* switch */

  /* allocate argc - optind peer entries */
  npeers = argc - optind;

  if (!(peers = (struct peer*)malloc(npeers * sizeof (struct peer)))) {
    fprintf(stderr, "malloc(): failed.\n");
    exit (1);
  }

  /* zero out malloc'd memory */
  bzero(peers, npeers*sizeof(struct peer));

  /* fill in peer entries */
  for (i = optind, n = 0; i < argc; ++i, ++n) {

    if (strlen(argv[i]) > 255) {
      fprintf(stderr, "ouch!\n");
      exit (1);
    }

    strcpy(tmp_buf, argv[i]);

    /* skip to end or port seperator */
    for (c = tmp_buf; (*c != '/') && (*c); ++c);

    /* extract the port part */
    if (*c == '/') {
      *c = 0;
      ++c;
      peers[n].addr.sin_port = atoi(c);
    } else 
      peers[n].addr.sin_port = DEFAULT_PORT;

    /* extract the ip address part */
    peers[n].addr.sin_addr.s_addr = scan_ip(tmp_buf);

    peers[n].addr.sin_family = AF_INET;

    if ((peers[n].fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
      fprintf(stderr, "socket(): %s\n", strerror(errno));
      exit (1);
    }
  }

  /* setup to receive flows */
  bzero(&fsocaddr, sizeof(fsocaddr));
  fsocaddr.sin_family = AF_INET;
  fsocaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  fsocaddr.sin_port = htons(fport);

  if ((fsockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    fprintf(stderr, "socket(): %s\n", strerror(errno));
    exit(1);
  }

  if (bind(fsockfd, (struct sockaddr*)&fsocaddr, sizeof(fsocaddr)) < 0) {
    fprintf(stderr, "bind(): %s\n", strerror(errno));
    exit (1);
  }

  len = sizeof(fsocaddr);

  while (1) {

    FD_ZERO (&rfd);
    FD_SET (fsockfd, &rfd);

    if (select (fsockfd+1, &rfd, (fd_set *)0, (fd_set *)0,
      (struct timeval*)0L) < 0)  {
      fprintf(stderr, "select(): %s\n", strerror(errno));
      exit (1);
    }

    if (FD_ISSET(fsockfd, &rfd)) {

      if ((n = recvfrom(fsockfd, &rcv_buf, sizeof rcv_buf, 0,
        (struct sockaddr*) &fsocaddr, &len)) == -1) {
        fprintf(stderr, "recvfrom(): %s\n", strerror(errno));
        exit (1);
      }

      for (i = 0; i < npeers; ++i) {

        sendto(peers[i].fd, &rcv_buf, n, 0, (struct sockaddr*)&peers[i].addr,
          sizeof (struct sockaddr_in));

        if (tx_delay)
          usleep((unsigned)tx_delay);

      }

    } /* if FD_ISSET */

  } /* while 1 */
} /* main */

void usage() {

  fprintf(stderr, "flow-fanout:\n\n");
  fprintf(stderr, " -d #  set debug level.\n");
  fprintf(stderr, " -p #  UDP port to accept flows on.\n");
  fprintf(stderr, " -x #  transmit delay in microseconds.\n");
  fprintf(stderr, " A.B.C.D/port  A.B.C.D/port ...\n");
  fprintf(stderr, "\n\n");

} /* usage */

